@@ -1,31 +1,6 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 Klarälvdalens Datakonsult AB, a KDAB Group company, info@kdab.com, author Milian Wolff <milian.wolff@kdab.com>
-** Copyright (C) 2019 Menlo Systems GmbH, author Arno Rehn <a.rehn@menlosystems.com>
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtWebChannel module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:GPL-EXCEPT$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 Klarälvdalens Datakonsult AB, a KDAB Group company, info@kdab.com, author Milian Wolff <milian.wolff@kdab.com>
+// Copyright (C) 2019 Menlo Systems GmbH, author Arno Rehn <a.rehn@menlosystems.com>
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 
 #include "tst_webchannel.h"
 
@@ -38,6 +13,13 @@
 #include <QJSEngine>
 #endif
 
+#include <QPromise>
+#include <QTimer>
+
+#ifdef WEBCHANNEL_TESTS_CAN_USE_CONCURRENT
+#include <QtConcurrent>
+#endif
+
 QT_USE_NAMESPACE
 
 #ifdef WEBCHANNEL_TESTS_CAN_USE_JS_ENGINE
@@ -183,6 +165,73 @@ void TestJSEngine::initWebChannelJS()
 
 #endif // WEBCHANNEL_TESTS_CAN_USE_JS_ENGINE
 
+namespace {
+QVariantList convert_to_js(const TestStructVector &list)
+{
+    QVariantList ret;
+    ret.reserve(list.size());
+    std::transform(list.begin(), list.end(), std::back_inserter(ret), [](const TestStruct &value) -> QVariant {
+        QVariantMap map;
+        map["foo"] = value.foo;
+        map["bar"] = value.bar;
+        return map;
+    });
+    return ret;
+}
+}
+
+#if QT_CONFIG(future)
+QFuture<int> TestObject::futureIntResult() const
+{
+    return QtFuture::makeReadyFuture(42);
+}
+
+QFuture<int> TestObject::futureDelayedIntResult() const
+{
+    QPromise<int> p;
+    const auto f = p.future();
+    p.start();
+    QTimer::singleShot(10, this, [p=std::move(p)]() mutable {
+        p.addResult(7);
+        p.finish();
+    });
+    return f;
+}
+
+#ifdef WEBCHANNEL_TESTS_CAN_USE_CONCURRENT
+QFuture<int> TestObject::futureIntResultFromThread() const
+{
+    return QtConcurrent::run([] {
+        return 1337;
+    });
+}
+#endif
+
+QFuture<void> TestObject::futureVoidResult() const
+{
+    return QtFuture::makeReadyFuture();
+}
+
+QFuture<QString> TestObject::futureStringResult() const
+{
+    return QtFuture::makeReadyFuture<QString>("foo");
+}
+
+QFuture<int> TestObject::cancelledFuture() const
+{
+    QPromise<int> p;
+    auto f = p.future();
+    p.start();
+    f.cancel();
+    Q_ASSERT(f.isCanceled());
+    return f;
+}
+
+QFuture<int> TestObject::failedFuture() const
+{
+    return QtFuture::makeExceptionalFuture<int>(QException{});
+}
+#endif
 
 TestWebChannel::TestWebChannel(QObject *parent)
     : QObject(parent)
@@ -191,6 +240,9 @@ TestWebChannel::TestWebChannel(QObject *parent)
     , m_lastBool(false)
     , m_lastDouble(0)
 {
+    qRegisterMetaType<TestStruct>();
+    qRegisterMetaType<TestStructVector>();
+    QMetaType::registerConverter<TestStructVector, QVariantList>(convert_to_js);
 }
 
 TestWebChannel::~TestWebChannel()
@@ -275,6 +327,21 @@ void TestWebChannel::setJsonArray(const QJsonArray& v)
     emit lastJsonArrayChanged();
 }
 
+int TestWebChannel::readOverload(int i)
+{
+    return i + 1;
+}
+
+QString TestWebChannel::readOverload(const QString &arg)
+{
+    return arg.toUpper();
+}
+
+QString TestWebChannel::readOverload(const QString &arg, int i)
+{
+    return arg.toUpper() + QString::number(i + 1);
+}
+
 void TestWebChannel::testRegisterObjects()
 {
     QWebChannel channel;
@@ -351,85 +418,58 @@ void TestWebChannel::testInfoForObject()
         QCOMPARE(info["enums"].toObject(), expected);
     }
 
+    QJsonArray expected;
+    auto addMethod = [&expected, &obj](const QString &name, const char *signature, bool addName = true) {
+        const auto index = obj.metaObject()->indexOfMethod(signature);
+        QVERIFY2(index != -1, signature);
+        if (addName)
+            expected.append(QJsonArray{name, index});
+        expected.append(QJsonArray{QString::fromUtf8(signature), index});
+    };
     { // methods & slots
-        QJsonArray expected;
-        {
-            QJsonArray method;
-            method.append(QStringLiteral("deleteLater"));
-            method.append(obj.metaObject()->indexOfMethod("deleteLater()"));
-            expected.append(method);
-        }
-        {
-            QJsonArray method;
-            method.append(QStringLiteral("slot1"));
-            method.append(obj.metaObject()->indexOfMethod("slot1()"));
-            expected.append(method);
-        }
-        {
-            QJsonArray method;
-            method.append(QStringLiteral("slot2"));
-            method.append(obj.metaObject()->indexOfMethod("slot2(QString)"));
-            expected.append(method);
-        }
-        {
-            QJsonArray method;
-            method.append(QStringLiteral("setReturnedObject"));
-            method.append(obj.metaObject()->indexOfMethod("setReturnedObject(TestObject*)"));
-            expected.append(method);
-        }
-        {
-            QJsonArray method;
-            method.append(QStringLiteral("setObjectProperty"));
-            method.append(obj.metaObject()->indexOfMethod("setObjectProperty(QObject*)"));
-            expected.append(method);
-        }
-        {
-            QJsonArray method;
-            method.append(QStringLiteral("setProp"));
-            method.append(obj.metaObject()->indexOfMethod("setProp(QString)"));
-            expected.append(method);
-        }
-        {
-            QJsonArray method;
-            method.append(QStringLiteral("fire"));
-            method.append(obj.metaObject()->indexOfMethod("fire()"));
-            expected.append(method);
-        }
-        {
-            QJsonArray method;
-            method.append(QStringLiteral("method1"));
-            method.append(obj.metaObject()->indexOfMethod("method1()"));
-            expected.append(method);
-        }
+        expected = {};
+        addMethod(QStringLiteral("deleteLater"), "deleteLater()");
+        addMethod(QStringLiteral("slot1"), "slot1()");
+        addMethod(QStringLiteral("slot2"), "slot2(QString)");
+        addMethod(QStringLiteral("setReturnedObject"), "setReturnedObject(TestObject*)");
+        addMethod(QStringLiteral("setObjectProperty"), "setObjectProperty(QObject*)");
+        addMethod(QStringLiteral("setProp"), "setProp(QString)");
+        addMethod(QStringLiteral("fire"), "fire()");
+        addMethod(QStringLiteral("overload"), "overload(double)");
+        addMethod(QStringLiteral("overload"), "overload(int)", false);
+        addMethod(QStringLiteral("overload"), "overload(QObject*)", false);
+        addMethod(QStringLiteral("overload"), "overload(QString)", false);
+        addMethod(QStringLiteral("overload"), "overload(QString,int)", false);
+        addMethod(QStringLiteral("overload"), "overload(QJsonArray)", false);
+        addMethod(QStringLiteral("setStringProperty"), "setStringProperty(QString)");
+        addMethod(QStringLiteral("bindableStringProperty"), "bindableStringProperty()");
+        addMethod(QStringLiteral("getStringProperty"), "getStringProperty()");
+        addMethod(QStringLiteral("bindStringPropertyToStringProperty2"), "bindStringPropertyToStringProperty2()");
+        addMethod(QStringLiteral("setStringProperty2"), "setStringProperty2(QString)");
+        addMethod(QStringLiteral("method1"), "method1()");
+#if QT_CONFIG(future)
+        addMethod(QStringLiteral("futureIntResult"), "futureIntResult()");
+        addMethod(QStringLiteral("futureDelayedIntResult"), "futureDelayedIntResult()");
+#ifdef WEBCHANNEL_TESTS_CAN_USE_CONCURRENT
+        addMethod(QStringLiteral("futureIntResultFromThread"), "futureIntResultFromThread()");
+#endif
+        addMethod(QStringLiteral("futureVoidResult"), "futureVoidResult()");
+        addMethod(QStringLiteral("futureStringResult"), "futureStringResult()");
+        addMethod(QStringLiteral("cancelledFuture"), "cancelledFuture()");
+        addMethod(QStringLiteral("failedFuture"), "failedFuture()");
+#endif
         QCOMPARE(info["methods"].toArray(), expected);
     }
 
     { // signals
-        QJsonArray expected;
-        {
-            QJsonArray signal;
-            signal.append(QStringLiteral("destroyed"));
-            signal.append(obj.metaObject()->indexOfMethod("destroyed(QObject*)"));
-            expected.append(signal);
-        }
-        {
-            QJsonArray signal;
-            signal.append(QStringLiteral("sig1"));
-            signal.append(obj.metaObject()->indexOfMethod("sig1()"));
-            expected.append(signal);
-        }
-        {
-            QJsonArray signal;
-            signal.append(QStringLiteral("sig2"));
-            signal.append(obj.metaObject()->indexOfMethod("sig2(QString)"));
-            expected.append(signal);
-        }
-        {
-            QJsonArray signal;
-            signal.append(QStringLiteral("replay"));
-            signal.append(obj.metaObject()->indexOfMethod("replay()"));
-            expected.append(signal);
-        }
+        expected = {};
+        addMethod(QStringLiteral("destroyed"), "destroyed(QObject*)");
+        addMethod(QStringLiteral("destroyed"), "destroyed()", false);
+        addMethod(QStringLiteral("sig1"), "sig1()");
+        addMethod(QStringLiteral("sig2"), "sig2(QString)");
+        addMethod(QStringLiteral("replay"), "replay()");
+        addMethod(QStringLiteral("overloadSignal"), "overloadSignal(int)");
+        addMethod(QStringLiteral("overloadSignal"), "overloadSignal(float)", false);
         QCOMPARE(info["signals"].toArray(), expected);
     }
 
@@ -524,6 +564,17 @@ void TestWebChannel::testInfoForObject()
             property.append(QJsonValue::fromVariant(QVariant::fromValue(obj.prop())));
             expected.append(property);
         }
+        {
+            QJsonArray property;
+            property.append(obj.metaObject()->indexOfProperty("stringProperty"));
+            property.append(QStringLiteral("stringProperty"));
+            {
+                QJsonArray signal;
+                property.append(signal);
+            }
+            property.append(QJsonValue::fromVariant(QVariant::fromValue(obj.readStringProperty())));
+            expected.append(property);
+        }
         QCOMPARE(info["properties"].toArray(), expected);
     }
 }
@@ -537,9 +588,7 @@ void TestWebChannel::testInvokeMethodConversion()
     args.append(QJsonValue(1000));
 
     {
-        int method = metaObject()->indexOfMethod("setInt(int)");
-        QVERIFY(method != -1);
-        channel.d_func()->publisher->invokeMethod(this, method, args);
+        channel.d_func()->publisher->invokeMethod(this, "setInt", args);
         QCOMPARE(m_lastInt, args.at(0).toInt());
         int getterMethod = metaObject()->indexOfMethod("readInt()");
         QVERIFY(getterMethod != -1);
@@ -547,11 +596,9 @@ void TestWebChannel::testInvokeMethodConversion()
         QCOMPARE(retVal, args.at(0).toVariant());
     }
     {
-        int method = metaObject()->indexOfMethod("setBool(bool)");
-        QVERIFY(method != -1);
         QJsonArray args;
         args.append(QJsonValue(!m_lastBool));
-        channel.d_func()->publisher->invokeMethod(this, method, args);
+        channel.d_func()->publisher->invokeMethod(this, "setBool", args);
         QCOMPARE(m_lastBool, args.at(0).toBool());
         int getterMethod = metaObject()->indexOfMethod("readBool()");
         QVERIFY(getterMethod != -1);
@@ -559,9 +606,7 @@ void TestWebChannel::testInvokeMethodConversion()
         QCOMPARE(retVal, args.at(0).toVariant());
     }
     {
-        int method = metaObject()->indexOfMethod("setDouble(double)");
-        QVERIFY(method != -1);
-        channel.d_func()->publisher->invokeMethod(this, method, args);
+        channel.d_func()->publisher->invokeMethod(this, "setDouble", args);
         QCOMPARE(m_lastDouble, args.at(0).toDouble());
         int getterMethod = metaObject()->indexOfMethod("readDouble()");
         QVERIFY(getterMethod != -1);
@@ -569,9 +614,7 @@ void TestWebChannel::testInvokeMethodConversion()
         QCOMPARE(retVal, args.at(0).toVariant());
     }
     {
-        int method = metaObject()->indexOfMethod("setVariant(QVariant)");
-        QVERIFY(method != -1);
-        channel.d_func()->publisher->invokeMethod(this, method, args);
+        channel.d_func()->publisher->invokeMethod(this, "setVariant", args);
         QCOMPARE(m_lastVariant, args.at(0).toVariant());
         int getterMethod = metaObject()->indexOfMethod("readVariant()");
         QVERIFY(getterMethod != -1);
@@ -579,23 +622,19 @@ void TestWebChannel::testInvokeMethodConversion()
         QCOMPARE(retVal, args.at(0).toVariant());
     }
     {
-        int method = metaObject()->indexOfMethod("setJsonValue(QJsonValue)");
-        QVERIFY(method != -1);
-        channel.d_func()->publisher->invokeMethod(this, method, args);
+        channel.d_func()->publisher->invokeMethod(this, "setJsonValue", args);
         QCOMPARE(m_lastJsonValue, args.at(0));
         int getterMethod = metaObject()->indexOfMethod("readJsonValue()");
         QVERIFY(getterMethod != -1);
         auto retVal = channel.d_func()->publisher->invokeMethod(this, getterMethod, {});
-        QCOMPARE(retVal, args.at(0).toVariant());
+        QCOMPARE(retVal, QVariant(args.at(0)));
     }
     {
-        int method = metaObject()->indexOfMethod("setJsonObject(QJsonObject)");
-        QVERIFY(method != -1);
         QJsonObject object;
         object["foo"] = QJsonValue(123);
         object["bar"] = QJsonValue(4.2);
         args[0] = object;
-        channel.d_func()->publisher->invokeMethod(this, method, args);
+        channel.d_func()->publisher->invokeMethod(this, "setJsonObject", args);
         QCOMPARE(m_lastJsonObject, object);
         int getterMethod = metaObject()->indexOfMethod("readJsonObject()");
         QVERIFY(getterMethod != -1);
@@ -603,13 +642,11 @@ void TestWebChannel::testInvokeMethodConversion()
         QCOMPARE(retVal, QVariant::fromValue(object));
     }
     {
-        int setterMethod = metaObject()->indexOfMethod("setJsonArray(QJsonArray)");
-        QVERIFY(setterMethod != -1);
         QJsonArray array;
         array << QJsonValue(123);
         array <<  QJsonValue(4.2);
         args[0] = array;
-        channel.d_func()->publisher->invokeMethod(this, setterMethod, args);
+        channel.d_func()->publisher->invokeMethod(this, "setJsonArray", args);
         QCOMPARE(m_lastJsonArray, array);
         int getterMethod = metaObject()->indexOfMethod("readJsonArray()");
         QVERIFY(getterMethod != -1);
@@ -618,6 +655,35 @@ void TestWebChannel::testInvokeMethodConversion()
     }
 }
 
+void TestWebChannel::testFunctionOverloading()
+{
+    QWebChannel channel;
+    channel.connectTo(m_dummyTransport);
+
+    // all method calls will use the first method's index
+    const auto method1 = metaObject()->indexOfMethod("readOverload(int)");
+    QVERIFY(method1 != -1);
+    const auto method2 = metaObject()->indexOfMethod("readOverload(QString)");
+    QVERIFY(method2 != -1);
+    QVERIFY(method1 < method2);
+    const auto method3 = metaObject()->indexOfMethod("readOverload(QString,int)");
+    QVERIFY(method3 != -1);
+    QVERIFY(method2 < method3);
+
+    { // int
+        const auto retVal = channel.d_func()->publisher->invokeMethod(this, method1, QJsonArray{1000});
+        QCOMPARE(retVal.toInt(), 1001);
+    }
+    { // QString
+        const auto retVal = channel.d_func()->publisher->invokeMethod(this, method2, QJsonArray{QStringLiteral("hello world")});
+        QCOMPARE(retVal.toString(), QStringLiteral("HELLO WORLD"));
+    }
+    { // QString, int
+        const auto retVal = channel.d_func()->publisher->invokeMethod(this, method3, QJsonArray{QStringLiteral("the answer is "), 41});
+        QCOMPARE(retVal.toString(), QStringLiteral("THE ANSWER IS 42"));
+    }
+}
+
 void TestWebChannel::testSetPropertyConversion()
 {
     QWebChannel channel;
@@ -676,6 +742,49 @@ void TestWebChannel::testSetPropertyConversion()
     }
 }
 
+void TestWebChannel::testInvokeMethodOverloadResolution()
+{
+    QWebChannel channel;
+    TestObject testObject;
+    TestObject exportedObject;
+    channel.registerObject("test", &exportedObject);
+    channel.connectTo(m_dummyTransport);
+
+    QVariant result;
+    QMetaObjectPublisher *publisher = channel.d_func()->publisher;
+
+    {
+        result = publisher->invokeMethod(&testObject, "overload", { 41.0 });
+        QVERIFY(result.userType() == QMetaType::Double);
+        QCOMPARE(result.toDouble(), 42.0);
+    }
+    {
+        // In JavaScript, there's only 'double', so this should always invoke the 'double' overload
+        result = publisher->invokeMethod(&testObject, "overload", { 41 });
+        QVERIFY(result.userType() == QMetaType::Double);
+        QCOMPARE(result.toDouble(), 42);
+    }
+    {
+        QJsonObject wrappedObject { {"id", "test"} };
+        result = publisher->invokeMethod(&testObject, "overload", { wrappedObject });
+        QCOMPARE(result.value<TestObject*>(), &exportedObject);
+    }
+    {
+        result = publisher->invokeMethod(&testObject, "overload", { "hello world" });
+        QCOMPARE(result.toString(), QStringLiteral("HELLO WORLD"));
+    }
+    {
+        result = publisher->invokeMethod(&testObject, "overload", { "the answer is ", 41 });
+        QCOMPARE(result.toString(), QStringLiteral("THE ANSWER IS 42"));
+    }
+    {
+        QJsonArray args;
+        args.append(QJsonArray { "foobar", 42 });
+        result = publisher->invokeMethod(&testObject, "overload", args);
+        QCOMPARE(result.toString(), QStringLiteral("42foobar"));
+    }
+}
+
 void TestWebChannel::testDisconnect()
 {
     QWebChannel channel;
@@ -727,6 +836,36 @@ void TestWebChannel::testUnwrapObject()
     }
 }
 
+void TestWebChannel::testTransportWrapObjectProperties()
+{
+    QWebChannel channel;
+
+    TestObject obj;
+    obj.setObjectName("testObject");
+    channel.registerObject(obj.objectName(), &obj);
+
+    DummyTransport *dummyTransport = new DummyTransport(this);
+    channel.connectTo(dummyTransport);
+    channel.d_func()->publisher->initializeClient(dummyTransport);
+    channel.d_func()->publisher->setClientIsIdle(true, dummyTransport);
+
+    QCOMPARE(channel.d_func()->publisher->transportedWrappedObjects.size(), 0);
+
+    QObject objPropObject;
+    objPropObject.setObjectName("foobar");
+
+    obj.setObjectProperty(&objPropObject);
+
+    channel.d_func()->publisher->sendPendingPropertyUpdates();
+
+    QCOMPARE(channel.d_func()->publisher->wrappedObjects.size(), 1);
+    const QString wrappedObjId = channel.d_func()->publisher->wrappedObjects.keys()[0];
+
+    QCOMPARE(channel.d_func()->publisher->transportedWrappedObjects.size(), 1);
+    QCOMPARE(channel.d_func()->publisher->transportedWrappedObjects.keys()[0], dummyTransport);
+    QCOMPARE(channel.d_func()->publisher->transportedWrappedObjects.values()[0], wrappedObjId);
+}
+
 void TestWebChannel::testRemoveUnusedTransports()
 {
     QWebChannel channel;
@@ -773,7 +912,7 @@ void TestWebChannel::testPassWrappedObjectBack()
     QJsonObject argProperty;
     argProperty["id"] = returnedObjPropertyInfo["id"];
 
-    pub->invokeMethod(&registeredObj, registeredObj.metaObject()->indexOfSlot("setReturnedObject(TestObject*)"), argsMethod);
+    pub->invokeMethod(&registeredObj, "setReturnedObject", argsMethod);
     QCOMPARE(registeredObj.mReturnedObject, &returnedObjMethod);
     pub->setProperty(&registeredObj, registeredObj.metaObject()->indexOfProperty("returnedObject"), argProperty);
     QCOMPARE(registeredObj.mReturnedObject, &returnedObjProperty);
@@ -798,12 +937,51 @@ void TestWebChannel::testWrapValues()
         QCOMPARE(value.toInt(), (int) flags);
     }
     {
-        QVector<int> vec{1, 2, 3};
-        QVariant variant = QVariant::fromValue(vec);
+        QList<int> list { 1, 2, 3 };
+        QVariant variant = QVariant::fromValue(list);
         QJsonValue value = channel.d_func()->publisher->wrapResult(variant, m_dummyTransport);
         QVERIFY(value.isArray());
         QCOMPARE(value.toArray(), QJsonArray({1, 2, 3}));
     }
+    {
+        TestStructVector vec{{1, 2}, {3, 4}};
+        QVariant variant = QVariant::fromValue(vec);
+        QJsonValue value = channel.d_func()->publisher->wrapResult(variant, m_dummyTransport);
+        QVERIFY(value.isArray());
+        QCOMPARE(value.toArray(), QJsonArray({QJsonObject{{"foo", 1}, {"bar", 2}},
+                                             QJsonObject{{"foo", 3}, {"bar", 4}}}));
+    }
+    {
+        QVariant variant = QVariant::fromValue(nullptr);
+        QJsonValue value = channel.d_func()->publisher->wrapResult(variant, m_dummyTransport);
+        QVERIFY(value.isNull());
+    }
+    {
+        QVariantHash hash;
+        hash["One"] = 1;
+        hash["Two"] = 2;
+        QVariant variant = QVariant::fromValue(hash);
+        QJsonValue value = channel.d_func()->publisher->wrapResult(variant, m_dummyTransport);
+        QVERIFY(value.isObject());
+        QVERIFY(value["One"].isDouble());
+        QCOMPARE(value["One"].toInt(), 1);
+        QVERIFY(value["Two"].isDouble());
+        QCOMPARE(value["Two"].toInt(), 2);
+        QVERIFY(value["Three"].isUndefined());
+    }
+    {
+        QVariantMap map;
+        map["One"] = 1;
+        map["Two"] = 2;
+        QVariant variant = QVariant::fromValue(map);
+        QJsonValue value = channel.d_func()->publisher->wrapResult(variant, m_dummyTransport);
+        QVERIFY(value.isObject());
+        QVERIFY(value["One"].isDouble());
+        QCOMPARE(value["One"].toInt(), 1);
+        QVERIFY(value["Two"].isDouble());
+        QCOMPARE(value["Two"].toInt(), 2);
+        QVERIFY(value["Three"].isUndefined());
+    }
 }
 
 void TestWebChannel::testWrapObjectWithMultipleTransports()
@@ -851,6 +1029,9 @@ void TestWebChannel::testInfiniteRecursion()
     channel.d_func()->publisher->initializeClient(m_dummyTransport);
 
     QJsonObject objectInfo = channel.d_func()->publisher->wrapResult(QVariant::fromValue(&obj), m_dummyTransport).toObject();
+
+    // Wrap the result twice to test for QTBUG-84007. A single result wrap will not trigger all recursion paths.
+    objectInfo = channel.d_func()->publisher->wrapResult(QVariant::fromValue(&obj), m_dummyTransport).toObject();
 }
 
 void TestWebChannel::testAsyncObject()
@@ -867,14 +1048,23 @@ void TestWebChannel::testAsyncObject()
     QJsonArray args;
     args.append(QJsonValue("message"));
 
-    int method = obj.metaObject()->indexOfMethod("setProp(QString)");
-    QVERIFY(method != -1);
+    {
+        int received = 0;
+        connect(&obj, &TestObject::propChanged, this, [&](const QString &arg) {
+            QCOMPARE(arg, args.at(0).toString());
+            ++received;
+        });
+        channel.d_func()->publisher->invokeMethod(&obj, "setProp", args);
+        QTRY_COMPARE(received, 1);
+    }
 
     {
-        QSignalSpy spy(&obj, &TestObject::propChanged);
-        channel.d_func()->publisher->invokeMethod(&obj, method, args);
-        QTRY_COMPARE(spy.count(), 1);
-        QCOMPARE(spy.at(0).at(0).toString(), args.at(0).toString());
+        int received = 0;
+        auto handler = obj.bindableStringProperty().onValueChanged([&] {
+            ++received;
+        });
+        channel.d_func()->publisher->invokeMethod(&obj, "setStringProperty", args);
+        QTRY_COMPARE(received, 1);
     }
 
     channel.registerObject("myObj", &obj);
@@ -887,18 +1077,232 @@ void TestWebChannel::testAsyncObject()
     channel.d_func()->publisher->handleMessage(connectMessage, m_dummyTransport);
 
     {
-        QSignalSpy spy(&obj, &TestObject::replay);
+        int received = 0;
+        connect(&obj, &TestObject::replay, this, [&]() { ++received; });
         QMetaObject::invokeMethod(&obj, "fire");
-        QTRY_COMPARE(spy.count(), 1);
+        QTRY_COMPARE(received, 1);
         channel.deregisterObject(&obj);
         QMetaObject::invokeMethod(&obj, "fire");
-        QTRY_COMPARE(spy.count(), 2);
+        QTRY_COMPARE(received, 2);
     }
 
     thread.quit();
     thread.wait();
 }
 
+void TestWebChannel::testQProperty()
+{
+    static const int IndexOfStringProperty =
+            TestObject::staticMetaObject.indexOfProperty("stringProperty");
+
+    DummyTransport transport;
+    QWebChannel channel;
+    QMetaObjectPublisher *publisher = channel.d_func()->publisher;
+
+    {
+        TestObject testObj;
+        testObj.setObjectName("testObject");
+
+        QProperty<QString> obj1("Hello");
+        testObj.bindableStringProperty().setBinding([&](){ return obj1.value(); });
+
+        QCOMPARE(obj1.value(), testObj.readStringProperty());
+
+        channel.registerObject(testObj.objectName(), &testObj);
+        channel.connectTo(&transport);
+
+        publisher->initializeClient(&transport);
+        // One bindable property should result in one observer
+        QCOMPARE(publisher->propertyObservers.count(&testObj), 1u);
+
+        QVariant result;
+        publisher->setClientIsIdle(true, &transport);
+        result = publisher->invokeMethod(&testObj, "getStringProperty", {});
+        QCOMPARE(result.toString(), obj1.value());
+
+        obj1 = "world";
+        result = publisher->invokeMethod(&testObj, "getStringProperty", {});
+        QCOMPARE(result.toString(), obj1.value());
+
+        publisher->sendPendingPropertyUpdates();
+        QVERIFY(!transport.messagesSent().isEmpty());
+        const QJsonObject updateMessage = transport.messagesSent().last()["data"][0].toObject();
+        QCOMPARE(updateMessage["object"], testObj.objectName());
+        QCOMPARE(updateMessage["properties"][QString::number(IndexOfStringProperty)], obj1.value());
+
+        publisher->invokeMethod(&testObj, "setStringProperty2", {"Hey"});
+        publisher->invokeMethod(&testObj, "bindStringPropertyToStringProperty2", {});
+        obj1 = "This should not affect getStringProperty";
+        result = publisher->invokeMethod(&testObj, "getStringProperty", {});
+        QCOMPARE(result.toString(), "Hey");
+
+        publisher->invokeMethod(&testObj, "setStringProperty2", {"again"});
+        result = publisher->invokeMethod(&testObj, "getStringProperty", {});
+        QCOMPARE(result.toString(), "again");
+    }
+
+    // Ensure that the observer has been removed after the object has been
+    // destroyed
+    QCOMPARE(publisher->propertyObservers.size(), 0u);
+}
+
+void TestWebChannel::testPropertyUpdateInterval_data()
+{
+    QTest::addColumn<int>("firstUpdateInterval");
+    QTest::addColumn<int>("firstUpdateWithin");
+    QTest::addColumn<int>("secondUpdateInterval");
+    QTest::addColumn<int>("secondUpdateWithin");
+
+    QTest::newRow("1500ms") << 1500 << 2000 << 1500 << 2000;
+    QTest::newRow("Next event") << 0 << 0 << 0 << 0;
+    QTest::newRow("Immediately") << -1 << 0 << -1 << 0;
+    QTest::newRow("1500ms then next event") << 1500 << 2000 << 0 << 0;
+    QTest::newRow("Next event then 1500ms") << 0 << 0 << 1500 << 2000;
+    QTest::newRow("Immediately the next event") << -1 << 0 << 0 << 0;
+    QTest::newRow("Next event then immediately") << 0 << 0 << -1 << 0;
+    QTest::newRow("Immediately then 1500ms") << -1 << 0 << 1500 << 2000;
+    QTest::newRow("1500ms then immediately") << 1500 << 2000 << -1 << 0;
+}
+
+void TestWebChannel::testPropertyUpdateInterval()
+{
+    DummyTransport transport;
+    QWebChannel channel;
+    QMetaObjectPublisher *publisher = channel.d_func()->publisher;
+
+    QFETCH(int, firstUpdateInterval);
+    QFETCH(int, firstUpdateWithin);
+
+    QProperty<int> propertyUpdateInterval(firstUpdateInterval);
+    channel.bindablePropertyUpdateInterval().setBinding(
+            Qt::makePropertyBinding(propertyUpdateInterval));
+    QCOMPARE(channel.propertyUpdateInterval(), firstUpdateInterval);
+
+    TestObject testObj;
+    testObj.setObjectName("testObject");
+    channel.registerObject(testObj.objectName(), &testObj);
+    channel.connectTo(&transport);
+    QProperty<QString> obj1("Hello");
+    testObj.bindableStringProperty().setBinding([&]() { return obj1.value(); });
+    publisher->initializeClient(&transport);
+    publisher->setClientIsIdle(true, &transport);
+    QVERIFY(transport.messagesSent().isEmpty());
+
+    // First update
+    obj1 = "world";
+    if (firstUpdateInterval > 0) {
+        QCoreApplication::processEvents();
+        QVERIFY(transport.messagesSent().isEmpty());
+        QTRY_COMPARE_WITH_TIMEOUT(transport.messagesSent().size(), 1u, firstUpdateWithin);
+    } else if (firstUpdateInterval == 0) {
+        QCOMPARE(transport.messagesSent().size(), 0u);
+        QCoreApplication::processEvents();
+        QCOMPARE(transport.messagesSent().size(), 1u);
+    } else {
+        QCOMPARE(transport.messagesSent().size(), 1u);
+    }
+
+    QFETCH(int, secondUpdateInterval);
+    QFETCH(int, secondUpdateWithin);
+    propertyUpdateInterval = secondUpdateInterval;
+    publisher->setClientIsIdle(true, &transport);
+    obj1 = "!!!";
+
+    if (secondUpdateInterval > 0) {
+        QCoreApplication::processEvents();
+        QCOMPARE(transport.messagesSent().size(), 1u);
+        QCOMPARE(channel.propertyUpdateInterval(), secondUpdateInterval);
+        QVERIFY(publisher->timer.isActive());
+        QTRY_COMPARE_WITH_TIMEOUT(transport.messagesSent().size(), 2u, secondUpdateWithin);
+    } else if (secondUpdateInterval == 0) {
+        QCOMPARE(transport.messagesSent().size(), 1u);
+        QCoreApplication::processEvents();
+        QCOMPARE(transport.messagesSent().size(), 2u);
+    } else {
+        QCOMPARE(transport.messagesSent().size(), 2u);
+    }
+}
+
+void TestWebChannel::testQPropertyBlockUpdates()
+{
+    DummyTransport transport;
+    QWebChannel channel;
+    QMetaObjectPublisher *publisher = channel.d_func()->publisher;
+    publisher->setPropertyUpdateInterval(0);
+    QProperty<bool> blockUpdates(false);
+    channel.bindableBlockUpdates().setBinding(Qt::makePropertyBinding(blockUpdates));
+    QCOMPARE(channel.blockUpdates(), false);
+    bool blockedSignalled(false);
+    connect(&channel, &QWebChannel::blockUpdatesChanged, this,
+            [&](bool block) { blockedSignalled = block; });
+
+    TestObject testObj;
+    testObj.setObjectName("testObject");
+    channel.registerObject(testObj.objectName(), &testObj);
+    channel.connectTo(&transport);
+    QProperty<QString> obj1("Hello");
+    testObj.bindableStringProperty().setBinding(Qt::makePropertyBinding(obj1));
+    publisher->initializeClient(&transport);
+    QVERIFY(transport.messagesSent().isEmpty());
+
+    blockUpdates = true;
+    QCOMPARE(blockedSignalled, true);
+    publisher->setClientIsIdle(true, &transport);
+    obj1 = "world";
+
+    QTest::qWait(0);
+    QCOMPARE(transport.messagesSent().size(), 0u);
+
+    blockUpdates = false;
+    QCOMPARE(blockedSignalled, false);
+    QCOMPARE(transport.messagesSent().size(), 1u);
+}
+
+void TestWebChannel::testPropertyMultipleTransports()
+{
+    DummyTransport transport1;
+    DummyTransport transport2;
+
+    QWebChannel channel;
+    QMetaObjectPublisher *publisher = channel.d_func()->publisher;
+
+    TestObject testObj;
+    testObj.setObjectName("testObject");
+    channel.registerObject(testObj.objectName(), &testObj);
+    channel.connectTo(&transport1);
+    channel.connectTo(&transport2);
+    QProperty<QString> obj1("Hello");
+    testObj.bindableStringProperty().setBinding([&]() { return obj1.value(); });
+
+    publisher->initializeClient(&transport1);
+    publisher->initializeClient(&transport2);
+    publisher->setClientIsIdle(true, &transport1);
+    QCOMPARE(publisher->isClientIdle(&transport1), true);
+    QCOMPARE(publisher->isClientIdle(&transport2), false);
+    channel.setPropertyUpdateInterval(1000);
+    QVERIFY(transport1.messagesSent().isEmpty());
+    QVERIFY(transport2.messagesSent().isEmpty());
+
+    obj1 = "World";
+    QTRY_COMPARE_WITH_TIMEOUT(transport1.messagesSent().size(), 1u, 2000);
+    QCOMPARE(transport2.messagesSent().size(), 0u);
+    publisher->setClientIsIdle(true, &transport2);
+    QTRY_COMPARE_WITH_TIMEOUT(transport2.messagesSent().size(), 1u, 2000);
+    QCOMPARE(publisher->isClientIdle(&transport1), false);
+    QCOMPARE(publisher->isClientIdle(&transport2), false);
+
+    obj1 = "!!!";
+    publisher->setClientIsIdle(true, &transport2);
+    QCOMPARE(publisher->isClientIdle(&transport2), true);
+    QCOMPARE(publisher->isClientIdle(&transport1), false);
+    QTRY_COMPARE_WITH_TIMEOUT(transport2.messagesSent().size(), 2u, 2000);
+    QCOMPARE(transport1.messagesSent().size(), 1u);
+    publisher->setClientIsIdle(true, &transport1);
+    QTRY_COMPARE_WITH_TIMEOUT(transport1.messagesSent().size(), 2u, 2000);
+    QCOMPARE(publisher->isClientIdle(&transport1), false);
+    QCOMPARE(publisher->isClientIdle(&transport2), false);
+}
+
 class FunctionWrapper : public QObject
 {
     Q_OBJECT
@@ -951,6 +1355,48 @@ void TestWebChannel::testDeletionDuringMethodInvocation()
         QCOMPARE(transport->messagesSent().size(), deleteChannel ? 0 : 1);
 }
 
+#if QT_CONFIG(future)
+void TestWebChannel::testAsyncMethodReturningFuture_data()
+{
+    QTest::addColumn<QString>("methodName");
+    QTest::addColumn<QJsonValue>("result");
+
+    QTest::addRow("int") << "futureIntResult" << QJsonValue{42};
+    QTest::addRow("int-delayed") << "futureDelayedIntResult" << QJsonValue{7};
+#ifdef WEBCHANNEL_TESTS_CAN_USE_CONCURRENT
+    QTest::addRow("int-thread") << "futureIntResultFromThread" << QJsonValue{1337};
+#endif
+    QTest::addRow("void") << "futureVoidResult" << QJsonValue{};
+    QTest::addRow("QString") << "futureStringResult" << QJsonValue{"foo"};
+
+    QTest::addRow("cancelled") << "cancelledFuture" << QJsonValue{};
+    QTest::addRow("failed")    << "failedFuture"    << QJsonValue{};
+}
+
+void TestWebChannel::testAsyncMethodReturningFuture()
+{
+    QFETCH(QString, methodName);
+    QFETCH(QJsonValue, result);
+
+    QWebChannel channel;
+    TestObject obj;
+    channel.registerObject("testObject", &obj);
+
+    DummyTransport transport;
+    channel.connectTo(&transport);
+
+    transport.emitMessageReceived({
+        {"type", TypeInvokeMethod},
+        {"object", "testObject"},
+        {"method", methodName},
+        {"id", 1}
+    });
+
+    QTRY_COMPARE(transport.messagesSent().size(), 1);
+    QCOMPARE(transport.messagesSent().first().value("data"), result);
+}
+#endif
+
 static QHash<QString, QObject*> createObjects(QObject *parent)
 {
     const int num = 100;
@@ -991,7 +1437,7 @@ void TestWebChannel::benchInitializeClients()
 
         publisher->propertyUpdatesInitialized = false;
         publisher->signalToPropertyMap.clear();
-        publisher->signalHandler.clear();
+        publisher->signalHandlers.clear();
     }
 }
 
@@ -1002,7 +1448,7 @@ void TestWebChannel::benchPropertyUpdates()
 
     QObject parent;
     const QHash<QString, QObject*> objects = createObjects(&parent);
-    QVector<BenchObject*> objectList;
+    QList<BenchObject *> objectList;
     objectList.reserve(objects.size());
     foreach (QObject *obj, objects) {
         objectList << qobject_cast<BenchObject*>(obj);
@@ -1016,7 +1462,7 @@ void TestWebChannel::benchPropertyUpdates()
             obj->change();
         }
 
-        channel.d_func()->publisher->clientIsIdle = true;
+        channel.d_func()->publisher->setClientIsIdle(true, m_dummyTransport);
         channel.d_func()->publisher->sendPendingPropertyUpdates();
     }
 }
@@ -1136,13 +1582,22 @@ void TestWebChannel::qtbug62388_wrapObjectMultipleTransports()
     auto queryObjectInfo = [&channel](QObject *obj, QWebChannelAbstractTransport *transport) {
         return channel.d_func()->publisher->wrapResult(QVariant::fromValue(obj), transport).toObject();
     };
-    const auto objectInfo = queryObjectInfo(&obj, m_dummyTransport);
 
-    QCOMPARE(objectInfo.length(), 3);
-    QVERIFY(objectInfo.contains("id"));
-    QVERIFY(objectInfo.contains("__QObject*__"));
-    QVERIFY(objectInfo.contains("data"));
-    QVERIFY(objectInfo.value("__QObject*__").isBool() && objectInfo.value("__QObject*__").toBool());
+    auto verifyObjectInfo = [&obj](const QJsonObject &objectInfo) {
+
+        QCOMPARE(objectInfo.length(), 3);
+        QVERIFY(objectInfo.contains("id"));
+        QVERIFY(objectInfo.contains("__QObject*__"));
+        QVERIFY(objectInfo.contains("data"));
+        QVERIFY(objectInfo.value("__QObject*__").isBool() && objectInfo.value("__QObject*__").toBool());
+
+        const auto propIndex = obj.metaObject()->indexOfProperty("prop");
+        const auto prop = objectInfo["data"].toObject()["properties"].toArray()[propIndex].toArray()[3].toString();
+        QCOMPARE(prop, obj.prop());
+    };
+
+    const auto objectInfo = queryObjectInfo(&obj, m_dummyTransport);
+    verifyObjectInfo(objectInfo);
 
     const auto id = objectInfo.value("id").toString();
 
@@ -1152,7 +1607,17 @@ void TestWebChannel::qtbug62388_wrapObjectMultipleTransports()
     initTransport(&transport);
     QCOMPARE(queryObjectInfo(&obj, &transport), objectInfo);
 
-    // don't crash when the transport is destroyed
+    obj.setProp("asdf");
+
+    const auto objectInfo2 = queryObjectInfo(&obj, m_dummyTransport);
+    QVERIFY(objectInfo2 != objectInfo);
+    verifyObjectInfo(objectInfo2);
+
+    DummyTransport transport2;
+    initTransport(&transport2);
+    QCOMPARE(queryObjectInfo(&obj, &transport2), objectInfo2);
+
+    // don't crash when the transports are destroyed
 }
 
 QTEST_MAIN(TestWebChannel)
